@@ -1,6 +1,6 @@
-/*	$Id: diff.c,v 1.41 2021/02/01 13:32:35 kristaps Exp $ */
+/*	$Id: diff.c,v 1.51 2021/02/17 13:31:36 kristaps Exp $ */
 /*
- * Copyright (c) 2017, 2018 Kristaps Dzonsons <kristaps@bsd.lv>
+ * Copyright (c) 2017--2021 Kristaps Dzonsons <kristaps@bsd.lv>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -37,26 +37,69 @@
 #include "libdiff.h"
 #include "extern.h"
 
+/*
+ * If "node" is not NULL, this represents our match attempts for a
+ * single node in a node tree.  We basically use "optmatch" and "opt" to
+ * keep trying to find the most optimal candidate in the other tree,
+ * which ends up being "match".
+ */
 struct	xnode {
-	char		 sig[MD5_DIGEST_STRING_LENGTH]; /* signature */
-	double		 weight; /* priority queue weight */
-	const struct lowdown_node *node; /* basis node */
-	const struct lowdown_node *match; /* matching node */
-	size_t		 opt; /* optimality of match */
-	const struct lowdown_node *optmatch; /* current optimal match */
+	char		 		 sig[MD5_DIGEST_STRING_LENGTH];
+	double		 		 weight; /* queue weight */
+	const struct lowdown_node 	*node; /* basis node */
+	const struct lowdown_node 	*match; /* matching node */
+	size_t		 		 opt; /* match optimality */
+	const struct lowdown_node 	*optmatch; /* current optimal */
 };
 
+/*
+ * A map of all nodes in the current tree by their ID.  A map can have
+ * holes (in which case the xnode's "node" is NULL) since we collapse
+ * adjacent text nodes as a preprocess.
+ */
 struct	xmap {
-	struct xnode	*nodes; /* all of the nodes (dense table) */
-	size_t		 maxsize; /* size of "nodes" (allocation) */
-	size_t		 maxid; /* maximum id in nodes */
-	size_t		 maxnodes; /* filled entries in "nodes" */
-	double		 maxweight; /* maximum node weight */
+	struct xnode	*nodes; /* holey table */
+	size_t		 maxsize; /* size of "nodes" */
+	size_t		 maxid; /* max node id */
+	size_t		 maxnodes; /* non-NULL count */
+	double		 maxweight; /* node weight */
 };
 
+/*
+ * Queue of nodes.  This is used in creating the priority queue of next
+ * nodes to parse.
+ */
 struct	pnode {
-	const struct lowdown_node *node; /* priority node */
-	TAILQ_ENTRY(pnode) entries;
+	const struct lowdown_node	*node; /* priority node */
+	TAILQ_ENTRY(pnode) 	 	 entries;
+};
+
+/*
+ * When we insert or delete reference (references never match), we
+ * might disrupt their order.  We want to maintain that the order of
+ * footnote reference is the order of their definitions.  Use this to
+ * re-order the definition queue.
+ */
+struct	ref {
+	size_t			 newnum; /* refnum of cloned */
+	size_t			 oldnum; /* original refnum */
+	enum lowdown_chng	 chng; /* from old or new tree */
+	TAILQ_ENTRY(ref)	 entries;
+};
+
+TAILQ_HEAD(refq, ref);
+
+/*
+ * Convenience structure to hold data we use when merging together the
+ * trees, mostly for the maps and reference queue for reassembling the
+ * ordered references.
+ */
+struct	merger {
+	const struct xmap *xoldmap; /* source xnodes */
+	const struct xmap *xnewmap; /* destination xnodes */
+	size_t		  id; /* maxid in new tree */
+	struct refq	  refq; /* ref re-id */
+	size_t		  refnum; /* next ref num to assign */
 };
 
 TAILQ_HEAD(pnodeq, pnode);
@@ -75,6 +118,7 @@ static void
 MD5Updatebuf(MD5_CTX *ctx, const struct lowdown_buf *v)
 {
 
+	assert(v != NULL);
 	MD5Update(ctx, (const uint8_t *)v->data, v->size);
 }
 
@@ -82,9 +126,24 @@ static void
 MD5Updatev(MD5_CTX *ctx, const void *v, size_t sz)
 {
 
+	assert(v != NULL);
 	MD5Update(ctx, (const unsigned char *)v, sz);
 }
 
+/*
+ * If this returns non-zero, the node should be considered opaque and
+ * we will not do any difference processing within it.  It will still be
+ * marked with weight and signature from child nodes and interior data.
+ */
+static int
+is_opaque(const struct lowdown_node *n)
+{
+
+	assert(n != NULL);
+	return n->type == LOWDOWN_TABLE_BLOCK ||
+		n->type == LOWDOWN_META;
+}
+
 /*
  * Assign signatures and weights.
  * This is defined by "Phase 2" in sec. 5.2., along with the specific
@@ -97,49 +156,57 @@ MD5Updatev(MD5_CTX *ctx, const void *v, size_t sz)
  */
 static double
 assign_sigs(MD5_CTX *parent, struct xmap *map, 
-	const struct lowdown_node *n)
+	const struct lowdown_node *n, int ign)
 {
 	const struct lowdown_node	*nn;
 	ssize_t				 weight = -1;
 	MD5_CTX				 ctx;
 	double				 v = 0.0, vv;
 	struct xnode			*xn;
+	struct xnode			 xntmp;
 	void				*pp;
+	int				 ign_chld = ign;
 
-	/* Get our node slot. */
+	/* 
+	 * Get our node slot unless we're ignoring the node.
+	 * Ignoring comes when a parent in our chain is opaque.
+	 */
 
-	if (n->id >= map->maxsize) {
-		pp = recallocarray(map->nodes, map->maxsize, 
-			n->id + 64, sizeof(struct xnode));
-		if (pp == NULL)
-			return -1.0;
-		map->nodes = pp;
-		map->maxsize = n->id + 64;
+	if (!ign) {
+		if (n->id >= map->maxsize) {
+			pp = recallocarray(map->nodes, map->maxsize, 
+				n->id + 64, sizeof(struct xnode));
+			if (pp == NULL)
+				return -1.0;
+			map->nodes = pp;
+			map->maxsize = n->id + 64;
+		}
+		xn = &map->nodes[n->id];
+		assert(xn->node == NULL);
+		assert(xn->weight == 0.0);
+		xn->node = n;
+		if (n->id > map->maxid)
+			map->maxid = n->id;
+		assert(map->maxid < map->maxsize);
+		map->maxnodes++;
+		ign_chld = is_opaque(n);
 	}
 
-	xn = &map->nodes[n->id];
-	assert(xn->node == NULL);
-	assert(xn->weight == 0.0);
-	xn->node = n;
-	if (n->id > map->maxid)
-		map->maxid = n->id;
-	assert(map->maxid < map->maxsize);
-	map->maxnodes++;
-
 	/* Recursive step. */
 
 	MD5Init(&ctx);
 	MD5Updatev(&ctx, &n->type, sizeof(enum lowdown_rndrt));
 
 	TAILQ_FOREACH(nn, &n->children, entries) {
-		if ((vv = assign_sigs(&ctx, map, nn)) < 0.0)
+		if ((vv = assign_sigs(&ctx, map, nn, ign_chld)) < 0.0)
 			return vv;
 		v += vv;
 	}
 
 	/* Re-assign "xn": child might have reallocated. */
 
-	xn = &map->nodes[n->id];
+	memset(&xntmp, 0, sizeof(struct xnode));
+	xn = ign ? &xntmp : &map->nodes[n->id];
 	xn->weight = v;
 
 	/*
@@ -162,6 +229,9 @@ assign_sigs(MD5_CTX *parent, struct xmap *map,
 	case LOWDOWN_CODESPAN:
 		weight = n->rndr_codespan.text.size;
 		break;
+	case LOWDOWN_META:
+		weight = n->rndr_meta.key.size;
+		break;
 	case LOWDOWN_IMAGE:
 		weight = n->rndr_image.link.size +
 			n->rndr_image.title.size +
@@ -213,6 +283,9 @@ assign_sigs(MD5_CTX *parent, struct xmap *map,
 	case LOWDOWN_NORMAL_TEXT:
 		MD5Updatebuf(&ctx, &n->rndr_normal_text.text);
 		break;
+	case LOWDOWN_META:
+		MD5Updatebuf(&ctx, &n->rndr_meta.key);
+		break;
 	case LOWDOWN_ENTITY:
 		MD5Updatebuf(&ctx, &n->rndr_entity.text);
 		break;
@@ -237,16 +310,15 @@ assign_sigs(MD5_CTX *parent, struct xmap *map,
 		MD5Updatebuf(&ctx, &n->rndr_codespan.text);
 		break;
 	case LOWDOWN_TABLE_HEADER:
-		/* Don't use the column metrics: mutable. */
+		MD5Updatev(&ctx, &n->rndr_table_header.columns,
+			sizeof(size_t));
 		break;
 	case LOWDOWN_TABLE_CELL:
 		MD5Updatev(&ctx, &n->rndr_table_cell.flags,
 			sizeof(enum htbl_flags));
-		/* Don't use the column number/count: mutable. */
+		MD5Updatev(&ctx, &n->rndr_table_cell.col,
+			sizeof(size_t));
 		break;
-	case LOWDOWN_FOOTNOTE_DEF:
-	case LOWDOWN_FOOTNOTE_REF:
-		/* Don't use footnote number: mutable. */
 	case LOWDOWN_IMAGE:
 		MD5Updatebuf(&ctx, &n->rndr_image.link);
 		MD5Updatebuf(&ctx, &n->rndr_image.title);
@@ -260,6 +332,13 @@ assign_sigs(MD5_CTX *parent, struct xmap *map,
 	case LOWDOWN_BLOCKHTML:
 		MD5Updatebuf(&ctx, &n->rndr_blockhtml.text);
 		break;
+	case LOWDOWN_FOOTNOTE_REF:
+		MD5Updatebuf(&ctx, &n->rndr_footnote_ref.key);
+		MD5Updatebuf(&ctx, &n->rndr_footnote_ref.def);
+		break;
+	case LOWDOWN_FOOTNOTE_DEF:
+		MD5Updatebuf(&ctx, &n->rndr_footnote_def.key);
+		break;
 	default:
 		break;
 	}
@@ -307,6 +386,7 @@ pqueue(const struct lowdown_node *n,
 		if (xnew->weight >= xold->weight)
 			break;
 	}
+
 	if (pp == NULL) {
 		TAILQ_INSERT_TAIL(pq, p, entries);
 		return 1;
@@ -320,6 +400,7 @@ pqueue(const struct lowdown_node *n,
 		if (p->node->id < pp->node->id)
 			break;
 	}
+
 	if (pp == NULL) 
 		TAILQ_INSERT_TAIL(pq, p, entries);
 	else
@@ -412,7 +493,6 @@ candidate(struct xnode *xnew, struct xmap *xnewmap,
  * Do the two internal nodes equal each other?
  * This depends upon the node type.
  * By default, all similarly-labelled (typed) nodes are equal.
- * We special-case as noted.
  */
 static int
 match_eq(const struct lowdown_node *n1, 
@@ -432,21 +512,18 @@ match_eq(const struct lowdown_node *n1,
 			return 0;
 		break;
 	case LOWDOWN_HEADER:
-		if (n1->rndr_header.level !=
-		    n2->rndr_header.level)
+		if (n1->rndr_header.level != n2->rndr_header.level)
 			return 0;
 		break;
-	case LOWDOWN_FOOTNOTE_DEF:
-		if (n1->rndr_footnote_def.num !=
-		    n2->rndr_footnote_def.num)
+	case LOWDOWN_META:
+		if (!hbuf_eq
+		    (&n1->rndr_meta.key, &n2->rndr_meta.key))
 			return 0;
 		break;
 	case LOWDOWN_LISTITEM:
-		if (n1->rndr_listitem.num !=
-		    n2->rndr_listitem.num)
+		if (n1->rndr_listitem.num != n2->rndr_listitem.num)
 			return 0;
-		if (n1->rndr_listitem.flags !=
-		    n2->rndr_listitem.flags)
+		if (n1->rndr_listitem.flags != n2->rndr_listitem.flags)
 			return 0;
 		break;
 	default:
@@ -560,6 +637,11 @@ match_down(struct xnode *xnew, struct xmap *xnewmap,
 	xnew->match = xold->node;
 	xold->match = xnew->node;
 
+	if (is_opaque(xnew->node)) {
+		assert(is_opaque(xold->node));
+		return;
+	}
+
 	nnew = TAILQ_FIRST(&xnew->node->children);
 	nold = TAILQ_FIRST(&xold->node->children);
 
@@ -587,6 +669,7 @@ node_clone(const struct lowdown_node *v, size_t id)
 {
 	struct lowdown_node	*n;
 	int			 rc = 1;
+	size_t			 i;
 
 	if ((n = calloc(1, sizeof(struct lowdown_node))) == NULL)
 		return NULL;
@@ -596,6 +679,10 @@ node_clone(const struct lowdown_node *v, size_t id)
 	n->id = id;
 
 	switch (n->type) {
+	case LOWDOWN_DEFINITION:
+		n->rndr_definition.flags =
+			v->rndr_definition.flags;
+		break;
 	case LOWDOWN_META:
 		rc = hbuf_clone(&v->rndr_meta.key, 
 			&n->rndr_meta.key);
@@ -645,17 +732,28 @@ node_clone(const struct lowdown_node *v, size_t id)
 		rc = hbuf_clone(&v->rndr_codespan.text,
 			&n->rndr_codespan.text);
 		break;
+	case LOWDOWN_TABLE_BLOCK:
+		n->rndr_table.columns = v->rndr_table.columns;
+		break;
 	case LOWDOWN_TABLE_HEADER:
-		/* Don't use the column metrics: mutable. */
+		n->rndr_table_header.columns = 
+			v->rndr_table_header.columns;
+		n->rndr_table_header.flags = calloc
+			(n->rndr_table_header.columns, 
+			 sizeof(enum htbl_flags));
+		if (n->rndr_table_header.flags == NULL)
+			return NULL;
+		for (i = 0; i < n->rndr_table_header.columns; i++)
+			n->rndr_table_header.flags[i] =
+				v->rndr_table_header.flags[i];
 		break;
 	case LOWDOWN_TABLE_CELL:
 		n->rndr_table_cell.flags = 
 			v->rndr_table_cell.flags;
-		/* Don't use the column number/count: mutable. */
-		break;
-	case LOWDOWN_FOOTNOTE_DEF:
-	case LOWDOWN_FOOTNOTE_REF:
-		/* Don't use footnote number: mutable. */
+		n->rndr_table_cell.col = 
+			v->rndr_table_cell.col;
+		n->rndr_table_cell.columns = 
+			v->rndr_table_cell.columns;
 		break;
 	case LOWDOWN_IMAGE:
 		rc = hbuf_clone(&v->rndr_image.link,
@@ -922,6 +1020,85 @@ out:
 	return rc;
 }
 
+/*
+ * Insert a footnote reference into our queue.
+ * This is because later we'll use ref_reorder() to make sure that the
+ * footnotes in the cloned tree are in the correct order.
+ */
+static int
+ref_insert(struct lowdown_node *nn, 
+	const struct lowdown_node *nf, struct merger *parms)
+{
+	struct ref	*ref;
+
+	if ((ref = calloc(1, sizeof(struct ref))) == NULL)
+		return 0;
+	assert(nf->type == LOWDOWN_FOOTNOTE_REF);
+	TAILQ_INSERT_TAIL(&parms->refq, ref, entries);
+	ref->oldnum = nf->rndr_footnote_ref.num;
+	nn->rndr_footnote_ref.num = parms->refnum++;
+	ref->newnum = nn->rndr_footnote_ref.num;
+	ref->chng = nn->chng;
+	return 1;
+}
+
+static void
+ref_free(struct merger *parms)
+{
+	struct ref	*ref;
+
+	while ((ref = TAILQ_FIRST(&parms->refq)) != NULL) {
+		TAILQ_REMOVE(&parms->refq, ref, entries);
+		free(ref);
+	}
+}
+
+/*
+ * Run this instead of trying to process the footnote block.
+ * This will manually go through all children of the footnote block
+ * (which must be footnote definitions, obviously) and put them in the
+ * correct order as dictated by the possibly-reordered ref_insert()
+ * entries.
+ */
+static int
+ref_reorder(struct lowdown_node *n, 
+	const struct lowdown_node *nold,
+	const struct lowdown_node *nnew,
+	struct merger *parms)
+{
+	struct ref			*ref;
+	const struct lowdown_node	*nn;
+	struct lowdown_node		*nc;
+
+	TAILQ_FOREACH(ref, &parms->refq, entries) {
+		nn = NULL;
+		if (ref->chng == LOWDOWN_CHNG_DELETE) {
+			TAILQ_FOREACH(nn, &nold->children, entries) {
+				assert(nn->type == LOWDOWN_FOOTNOTE_DEF);
+				if (nn->rndr_footnote_def.num == ref->oldnum)
+					break;
+			}
+		} else {
+			TAILQ_FOREACH(nn, &nnew->children, entries) {
+				assert(nn->type == LOWDOWN_FOOTNOTE_DEF);
+				if (nn->rndr_footnote_def.num == ref->oldnum)
+					break;
+			}
+		}
+		if (nn == NULL)
+			continue;
+		if ((nc = node_clonetree(nn, &parms->id)) == NULL)
+			return 0;
+		TAILQ_INSERT_TAIL(&n->children, nc, entries);
+		nc->parent = n;
+		nc->chng = ref->chng;
+		assert(nc->type == LOWDOWN_FOOTNOTE_DEF);
+		nc->rndr_footnote_def.num = ref->newnum;
+	}
+
+	return 1;
+}
+
 /*
  * Merge the new tree "nnew" with the old "nold" using a depth-first
  * algorithm.
@@ -933,14 +1110,13 @@ out:
  */
 static struct lowdown_node *
 node_merge(const struct lowdown_node *nold,
-	const struct xmap *xoldmap,
-	const struct lowdown_node *nnew,
-	const struct xmap *xnewmap,
-	size_t *id)
+	const struct lowdown_node *nnew, struct merger *parms)
 {
 	const struct xnode		*xnew, *xold;
 	struct lowdown_node		*n, *nn;
 	const struct lowdown_node	*nnold;
+	const struct xmap 		*xoldmap = parms->xoldmap,
+	      				*xnewmap = parms->xnewmap;
 
 	/* 
 	 * Invariant: the current nodes are matched.
@@ -954,9 +1130,22 @@ node_merge(const struct lowdown_node *nold,
 	assert(xold->match != NULL);
 	assert(xnew->match == xold->node);
 
-	if ((n = node_clone(nnew, (*id)++)) == NULL)
+	if ((n = node_clone(nnew, parms->id++)) == NULL)
 		goto err;
 
+	/* 
+	 * Special handling of footnotes: we might have rearranged the
+	 * order in which footnote definitions are displayed.  So now we
+	 * go through the queue of footnotes and insert them in the
+	 * order they were defined.
+	 */
+
+	if (n->type == LOWDOWN_FOOTNOTES_BLOCK) {
+		if (!ref_reorder(n, nold, nnew, parms))
+			goto err;
+		return n;
+	}
+
 	/* Now walk through the children on both sides. */
 
 	nold = TAILQ_FIRST(&nold->children);
@@ -976,11 +1165,15 @@ node_merge(const struct lowdown_node *nold,
 			if (xold->match != NULL ||
 			    LOWDOWN_NORMAL_TEXT == nold->type)
 				break;
-			if ((nn = node_clonetree(nold, id)) == NULL)
+			if ((nn = node_clonetree
+			    (nold, &parms->id)) == NULL)
 				goto err;
 			TAILQ_INSERT_TAIL(&n->children, nn, entries);
 			nn->parent = n;
 			nn->chng = LOWDOWN_CHNG_DELETE;
+			if (nn->type == LOWDOWN_FOOTNOTE_REF &&
+			    !ref_insert(nn, nold, parms))
+				goto err;
 			nold = TAILQ_NEXT(nold, entries);
 		}
 
@@ -996,11 +1189,15 @@ node_merge(const struct lowdown_node *nold,
 			if (xnew->match != NULL ||
 			    LOWDOWN_NORMAL_TEXT == nnew->type)
 				break;
-			if ((nn = node_clonetree(nnew, id)) == NULL)
+			if ((nn = node_clonetree
+			    (nnew, &parms->id)) == NULL)
 				goto err;
 			TAILQ_INSERT_TAIL(&n->children, nn, entries);
 			nn->parent = n;
 			nn->chng = LOWDOWN_CHNG_INSERT;
+			if (nn->type == LOWDOWN_FOOTNOTE_REF &&
+			    !ref_insert(nn, nnew, parms))
+				goto err;
 			nnew = TAILQ_NEXT(nnew, entries);
 		}
 
@@ -1015,7 +1212,7 @@ node_merge(const struct lowdown_node *nold,
 		    xold->match == NULL &&
 		    nnew->type == LOWDOWN_NORMAL_TEXT &&
 		    xnew->match == NULL) {
-			if (!node_lcs(nold, nnew, n, id))
+			if (!node_lcs(nold, nnew, n, &parms->id))
 				goto err;
 			nold = TAILQ_NEXT(nold, entries);
 			nnew = TAILQ_NEXT(nnew, entries);
@@ -1025,23 +1222,31 @@ node_merge(const struct lowdown_node *nold,
 			xold = &xoldmap->nodes[nold->id];
 			if (xold->match != NULL)
 				break;
-			if ((nn = node_clonetree(nold, id)) == NULL)
+			if ((nn = node_clonetree
+			    (nold, &parms->id)) == NULL)
 				goto err;
 			TAILQ_INSERT_TAIL(&n->children, nn, entries);
 			nn->parent = n;
 			nn->chng = LOWDOWN_CHNG_DELETE;
+			if (nn->type == LOWDOWN_FOOTNOTE_REF &&
+			    !ref_insert(nn, nold, parms))
+				goto err;
 			nold = TAILQ_NEXT(nold, entries);
 		}
 
-		while ( nnew != NULL) {
+		while (nnew != NULL) {
 			xnew = &xnewmap->nodes[nnew->id];
 			if (xnew->match != NULL)
 				break;
-			if ((nn = node_clonetree(nnew, id)) == NULL)
+			if ((nn = node_clonetree
+			    (nnew, &parms->id)) == NULL)
 				goto err;
 			TAILQ_INSERT_TAIL(&n->children, nn, entries);
 			nn->parent = n;
 			nn->chng = LOWDOWN_CHNG_INSERT;
+			if (nn->type == LOWDOWN_FOOTNOTE_REF &&
+			    !ref_insert(nn, nnew, parms))
+				goto err;
 			nnew = TAILQ_NEXT(nnew, entries);
 		}
 
@@ -1078,11 +1283,15 @@ node_merge(const struct lowdown_node *nold,
 		 */
 
 		if (nnold == NULL) {
-			if ((nn = node_clonetree(nnew, id)) == NULL)
+			if ((nn = node_clonetree
+			    (nnew, &parms->id)) == NULL)
 				goto err;
 			TAILQ_INSERT_TAIL(&n->children, nn, entries);
 			nn->parent = n;
 			nn->chng = LOWDOWN_CHNG_INSERT;
+			if (nn->type == LOWDOWN_FOOTNOTE_REF &&
+			    !ref_insert(nn, nnew, parms))
+				goto err;
 			nnew = TAILQ_NEXT(nnew, entries);
 			continue;
 		}
@@ -1093,11 +1302,15 @@ node_merge(const struct lowdown_node *nold,
 			xold = &xoldmap->nodes[nold->id];
 			if (xnew->node == xold->match) 
 				break;
-			if ((nn = node_clonetree(nold, id)) == NULL)
+			if ((nn = node_clonetree
+			    (nold, &parms->id)) == NULL)
 				goto err;
 			TAILQ_INSERT_TAIL(&n->children, nn, entries);
 			nn->parent = n;
 			nn->chng = LOWDOWN_CHNG_DELETE;
+			if (nn->type == LOWDOWN_FOOTNOTE_REF &&
+			    !ref_insert(nn, nold, parms))
+				goto err;
 			nold = TAILQ_NEXT(nold, entries);
 		}
 
@@ -1109,11 +1322,24 @@ node_merge(const struct lowdown_node *nold,
 		 * Then continue on to the next nodes.
 		 */
 
-		nn = node_merge(nold, xoldmap, nnew, xnewmap, id);
-		if (nn == NULL)
-			goto err;
-		TAILQ_INSERT_TAIL(&n->children, nn, entries);
-		nn->parent = n;
+		if (is_opaque(nnew)) {
+			assert(is_opaque(nold));
+			if ((nn = node_clonetree
+			    (nnew, &parms->id)) == NULL)
+				goto err;
+			TAILQ_INSERT_TAIL(&n->children, nn, entries);
+			nn->parent = n;
+		} else {
+			assert(!is_opaque(nold));
+			nn = node_merge(nold, nnew, parms);
+			if (nn == NULL)
+				goto err;
+			TAILQ_INSERT_TAIL(&n->children, nn, entries);
+			nn->parent = n;
+			if (nn->type == LOWDOWN_FOOTNOTE_REF &&
+			    !ref_insert(nn, nnew, parms))
+				goto err;
+		}
 
 		nold = TAILQ_NEXT(nold, entries);
 		nnew = TAILQ_NEXT(nnew, entries);
@@ -1122,11 +1348,14 @@ node_merge(const struct lowdown_node *nold,
 	/* Flush remaining old nodes. */
 
 	while (nold != NULL) {
-		if ((nn = node_clonetree(nold, id)) == NULL)
+		if ((nn = node_clonetree (nold, &parms->id)) == NULL)
 			goto err;
 		TAILQ_INSERT_TAIL(&n->children, nn, entries);
 		nn->parent = n;
 		nn->chng = LOWDOWN_CHNG_DELETE;
+		if (nn->type == LOWDOWN_FOOTNOTE_REF &&
+		    !ref_insert(nn, nold, parms))
+			goto err;
 		nold = TAILQ_NEXT(nold, entries);
 	}
 
@@ -1152,7 +1381,7 @@ node_optimise_topdown(const struct lowdown_node *n,
 	const struct lowdown_node	*match, *nchild, *mchild, 
 	      				*nnext, *mnext;
 
-	if (TAILQ_EMPTY(&n->children))
+	if (is_opaque(n) || TAILQ_EMPTY(&n->children))
 		return;
 
 	xn = &newmap->nodes[n->id];
@@ -1165,7 +1394,7 @@ node_optimise_topdown(const struct lowdown_node *n,
 	assert(xmatch != NULL);
 
 	TAILQ_FOREACH(nchild, &n->children, entries) {
-		if (TAILQ_EMPTY(&nchild->children))
+		if (is_opaque(nchild) || TAILQ_EMPTY(&nchild->children))
 			continue;
 		xnchild = &newmap->nodes[nchild->id];
 		assert(xnchild != NULL);
@@ -1183,7 +1412,7 @@ node_optimise_topdown(const struct lowdown_node *n,
 
 		if ((nnext = TAILQ_NEXT(nchild, entries)) == NULL)
 			continue;
-		if (TAILQ_EMPTY(&nnext->children))
+		if (is_opaque(nnext) || TAILQ_EMPTY(&nnext->children))
 			continue;
 		xnnext = &newmap->nodes[nnext->id];
 		assert(xnnext != NULL);
@@ -1192,14 +1421,14 @@ node_optimise_topdown(const struct lowdown_node *n,
 
 		if ((mnext = TAILQ_NEXT(mchild, entries)) == NULL)
 			continue;
-		if (TAILQ_EMPTY(&mnext->children))
+		if (is_opaque(mnext) || TAILQ_EMPTY(&mnext->children))
 			continue;
 		xmnext = &oldmap->nodes[mnext->id];
 		assert(xmnext != NULL);
 		if (xmnext->match != NULL)
 			continue;
 
-		if ( ! match_eq(nnext, mnext))
+		if (!match_eq(nnext, mnext))
 			continue;
 
 		xnnext->match = mnext;
@@ -1226,11 +1455,13 @@ node_optimise_bottomup(const struct lowdown_node *n,
 	const struct lowdown_node	*nn, *on, *nnn, *maxn = NULL;
 	double				 w, maxw = 0.0, tw = 0.0;
 
-	/* Do a depth-first pre-order search. */
+	/* Ignore opaque nodes. */
 
-	if (TAILQ_EMPTY(&n->children))
+	if (is_opaque(n) || TAILQ_EMPTY(&n->children))
 		return;
 
+	/* Do a depth-first pre-order search. */
+
 	TAILQ_FOREACH(nn, &n->children, entries) {
 		tw += newmap->nodes[nn->id].weight;
 		node_optimise_bottomup(nn, newmap, oldmap);
@@ -1308,6 +1539,7 @@ lowdown_diff(const struct lowdown_node *nold,
 	const struct lowdown_node	*n, *nn;
 	struct lowdown_node		*comp = NULL;
 	size_t				 i;
+	struct merger			 parms;
 
 	memset(&xoldmap, 0, sizeof(struct xmap));
 	memset(&xnewmap, 0, sizeof(struct xmap));
@@ -1319,9 +1551,9 @@ lowdown_diff(const struct lowdown_node *nold,
 	 * See "Phase 2", sec 5.2.
 	 */
 
-	if (assign_sigs(NULL, &xoldmap, nold) < 0.0)
+	if (assign_sigs(NULL, &xoldmap, nold, 0) < 0.0)
 		goto out;
-	if (assign_sigs(NULL, &xnewmap, nnew) < 0.0)
+	if (assign_sigs(NULL, &xnewmap, nnew, 0) < 0.0)
 		goto out;
 
 	/* Prime the priority queue with the root. */
@@ -1365,9 +1597,14 @@ lowdown_diff(const struct lowdown_node *nold,
 			candidate(xnew, &xnewmap, xold, &xoldmap);
 		}
 
-		/* No match: enqueue children ("Phase 3" cont.). */
+		/* 
+		 * No match: enqueue children ("Phase 3" cont.).
+		 * Ignore opaque nodes.
+		 */
 
 		if (xnew->optmatch == NULL) {
+			if (is_opaque(n))
+				continue;
 			TAILQ_FOREACH(nn, &n->children, entries)
 				if (!pqueue(nn, &xnewmap, &pq))
 					goto out;
@@ -1405,8 +1642,12 @@ lowdown_diff(const struct lowdown_node *nold,
 	 * See "Phase 5", sec. 5.2.
 	 */
 
-	i = 0;
-	comp = node_merge(nold, &xoldmap, nnew, &xnewmap, &i);
+	memset(&parms, 0, sizeof(struct merger));
+	parms.xoldmap = &xoldmap;
+	parms.xnewmap = &xnewmap;
+	TAILQ_INIT(&parms.refq);
+	comp = node_merge(nold, nnew, &parms);
+	ref_free(&parms);
 
 	*maxn = xnewmap.maxid > xoldmap.maxid ?
 		xnewmap.maxid + 1 :
