@@ -1,4 +1,4 @@
-/*	$Id: html.c,v 1.100 2021/02/01 13:18:56 kristaps Exp $ */
+/*	$Id: html.c,v 1.102 2021/02/16 23:15:41 kristaps Exp $ */
 /*
  * Copyright (c) 2008, Natacha Porté
  * Copyright (c) 2011, Vicent Martí
@@ -857,7 +857,7 @@ rndr_footnote_def(struct lowdown_buf *ob,
 		break;
 	}
 
-	if (!hbuf_printf(ob, "\n<li id=\"fn%zu\">\n", param->num))
+	if (!hbuf_printf(ob, "\n<li id=\"fn%zu\">\n", param->num + 1))
 		return 0;
 
 	if (pfound) {
@@ -866,7 +866,7 @@ rndr_footnote_def(struct lowdown_buf *ob,
 		if (!hbuf_printf(ob, "&#160;"
 		    "<a href=\"#fnref%zu\" rev=\"footnote\">"
 		    "&#8617;"
-		    "</a>", param->num))
+		    "</a>", param->num + 1))
 			return 0;
 		if (!hbuf_put(ob, 
 		    content->data + i, content->size - i))
@@ -888,7 +888,7 @@ rndr_footnote_ref(struct lowdown_buf *ob,
 		"<sup id=\"fnref%zu\">"
 		"<a href=\"#fn%zu\" rel=\"footnote\">"
 		"%zu</a></sup>", 
-		param->num, param->num, param->num);
+		param->num + 1, param->num + 1, param->num + 1);
 }
 
 static int
@@ -1121,7 +1121,7 @@ rndr(struct lowdown_buf *ob,
 	struct lowdown_buf		*tmp;
 	int32_t				 ent;
 	struct html			*st = ref;
-	int				 ret = 1, rc = 0;
+	int				 ret = 1, rc = 1;
 
 	if ((tmp = hbuf_new(64)) == NULL)
 		return 0;
@@ -1160,7 +1160,8 @@ rndr(struct lowdown_buf *ob,
 		rc = rndr_doc_header(ob, tmp, mq, st);
 		break;
 	case LOWDOWN_META:
-		rc = rndr_meta(ob, tmp, mq, n, st);
+		if (n->chng != LOWDOWN_CHNG_DELETE)
+			rc = rndr_meta(ob, tmp, mq, n, st);
 		break;
 	case LOWDOWN_DOC_FOOTER:
 		rc = rndr_doc_footer(ob, st);
