@@ -1,4 +1,4 @@
-/*	$Id: latex.c,v 1.21 2021/02/01 13:18:56 kristaps Exp $ */
+/*	$Id: latex.c,v 1.24 2021/02/16 23:26:35 kristaps Exp $ */
 /*
  * Copyright (c) 2020--2021 Kristaps Dzonsons <kristaps@bsd.lv>
  *
@@ -593,13 +593,24 @@ rndr_normal_text(struct lowdown_buf *ob,
 static int
 rndr_footnote_def(struct lowdown_buf *ob,
 	const struct lowdown_buf *content, 
+	const struct lowdown_node *n,
 	const struct rndr_footnote_def *param)
 {
 
-	if (!hbuf_printf(ob, "\\footnotetext[%zu]{", param->num))
+	if (!hbuf_printf(ob, "\\footnotetext[%zu]{", param->num + 1))
+		return 0;
+	if (n->chng == LOWDOWN_CHNG_INSERT &&
+	    !HBUF_PUTSL(ob, "\\textcolor{blue}{"))
+		return 0;
+	if (n->chng == LOWDOWN_CHNG_DELETE &&
+	    !HBUF_PUTSL(ob, "\\textcolor{red}{"))
 		return 0;
 	if (!hbuf_putb(ob, content))
 		return 0;
+	if ((n->chng == LOWDOWN_CHNG_INSERT ||
+	     n->chng == LOWDOWN_CHNG_DELETE) &&
+	    !HBUF_PUTSL(ob, "}"))
+		return 0;
 	return HBUF_PUTSL(ob, "}\n");
 }
 
@@ -608,7 +619,7 @@ rndr_footnote_ref(struct lowdown_buf *ob,
 	const struct rndr_footnote_ref *param)
 {
 
-	return hbuf_printf(ob, "\\footnotemark[%zu]", param->num);
+	return hbuf_printf(ob, "\\footnotemark[%zu]", param->num + 1);
 }
 
 static int
@@ -743,7 +754,7 @@ rndr(struct lowdown_buf *ob,
 	struct lowdown_buf		*tmp;
 	struct latex			*st = arg;
 	const struct lowdown_node	*child;
-	int				 ret = 0, rc;
+	int				 ret = 0, rc = 1;
 
 	if ((tmp = hbuf_new(64)) == NULL)
 		return 0;
@@ -758,9 +769,11 @@ rndr(struct lowdown_buf *ob,
 	 */
 
 	if (n->chng == LOWDOWN_CHNG_INSERT && 
+	    n->type != LOWDOWN_FOOTNOTE_DEF &&
 	    !HBUF_PUTSL(ob, "{\\color{blue} "))
 		goto out;
 	if (n->chng == LOWDOWN_CHNG_DELETE &&
+	    n->type != LOWDOWN_FOOTNOTE_DEF &&
 	    !HBUF_PUTSL(ob, "{\\color{red} "))
 		goto out;
 
@@ -781,7 +794,8 @@ rndr(struct lowdown_buf *ob,
 		rc = rndr_doc_header(ob, mq, st);
 		break;
 	case LOWDOWN_META:
-		rc = rndr_meta(ob, tmp, mq, n, st);
+		if (n->chng != LOWDOWN_CHNG_DELETE)
+			rc = rndr_meta(ob, tmp, mq, n, st);
 		break;
 	case LOWDOWN_DOC_FOOTER:
 		rc = rndr_doc_footer(ob, st);
@@ -811,7 +825,8 @@ rndr(struct lowdown_buf *ob,
 		rc = rndr_tablecell(ob, tmp, &n->rndr_table_cell);
 		break;
 	case LOWDOWN_FOOTNOTE_DEF:
-		rc = rndr_footnote_def(ob, tmp, &n->rndr_footnote_def);
+		rc = rndr_footnote_def
+			(ob, tmp, n, &n->rndr_footnote_def);
 		break;
 	case LOWDOWN_BLOCKHTML:
 		rc = rndr_raw_block(ob, &n->rndr_blockhtml, st);
@@ -868,10 +883,11 @@ rndr(struct lowdown_buf *ob,
 	if (!rc)
 		goto out;
 
-	if (n->chng == LOWDOWN_CHNG_INSERT ||
-	    n->chng == LOWDOWN_CHNG_DELETE)
-		if (!HBUF_PUTSL(ob, "}"))
-			goto out;
+	if ((n->chng == LOWDOWN_CHNG_INSERT ||
+	     n->chng == LOWDOWN_CHNG_DELETE) &&
+	    n->type != LOWDOWN_FOOTNOTE_DEF &&
+	    !HBUF_PUTSL(ob, "}"))
+		goto out;
 
 	ret = 1;
 out:
