@@ -1,4 +1,4 @@
-/*	$Id: main.c,v 1.98 2021/02/08 20:10:54 kristaps Exp $ */
+/*	$Id: main.c,v 1.102 2021/02/18 22:37:50 kristaps Exp $ */
 /*
  * Copyright (c) 2016, 2017, 2020 Kristaps Dzonsons <kristaps@bsd.lv>
  *
@@ -208,7 +208,9 @@ main(int argc, char *argv[])
 	FILE			*fin = stdin, *fout = stdout, 
 				*din = NULL;
 	const char		*fnin = "<stdin>", *fnout = NULL,
-	      	 		*fndin = NULL, *extract = NULL, *er;
+	      	 		*fndin = NULL, *extract = NULL, *er,
+				*mainopts = "M:m:sT:o:X:",
+	      			*diffopts = "M:m:sT:o:";
 	struct lowdown_opts 	 opts;
 	int			 c, diff = 0,
 				 status = 0, aoflag = 0, roflag = 0,
@@ -250,6 +252,10 @@ main(int argc, char *argv[])
 		{ "term-columns",	required_argument, NULL, 4 },
 		{ "gemini-link-end",	no_argument, 	&aoflag, LOWDOWN_GEMINI_LINK_END },
 		{ "gemini-no-link-end",	no_argument, 	&roflag, LOWDOWN_GEMINI_LINK_END },
+		{ "gemini-link-roman",	no_argument, 	&aoflag, LOWDOWN_GEMINI_LINK_ROMAN },
+		{ "gemini-no-link-roman", no_argument, 	&roflag, LOWDOWN_GEMINI_LINK_ROMAN },
+		{ "gemini-link-noref",	no_argument, 	&aoflag, LOWDOWN_GEMINI_LINK_NOREF },
+		{ "gemini-no-link-noref", no_argument, 	&roflag, LOWDOWN_GEMINI_LINK_NOREF },
 		{ "gemini-link-inline",	no_argument, 	&aoflag, LOWDOWN_GEMINI_LINK_IN },
 		{ "gemini-no-link-inline",no_argument, 	&roflag, LOWDOWN_GEMINI_LINK_IN },
 		{ "gemini-metadata",	no_argument, 	&aoflag, LOWDOWN_GEMINI_METADATA },
@@ -331,8 +337,8 @@ main(int argc, char *argv[])
 	if (strcasecmp(getprogname(), "lowdown-diff") == 0) 
 		diff = 1;
 
-	while ((c = getopt_long
-		(argc, argv, "M:m:sT:o:X:", lo, NULL)) != -1)
+	while ((c = getopt_long(argc, argv, 
+	       diff ? diffopts : mainopts, lo, NULL)) != -1)
 		switch (c) {
 		case 'M':
 			metadata_parse(c, &opts.metaovr, 
@@ -419,7 +425,8 @@ main(int argc, char *argv[])
 	argc -= optind;
 	argv += optind;
 
-	if (opts.type == LOWDOWN_TERM)
+	if (opts.type == LOWDOWN_TERM ||
+ 	    opts.type == LOWDOWN_GEMINI)
 		setlocale(LC_CTYPE, "");
 
 	/* 
@@ -525,30 +532,13 @@ main(int argc, char *argv[])
 	lowdown_metaq_free(&mq);
 	return status;
 usage:
-	fprintf(stderr, "usage: lowdown "
-		"[-s] "
-		"[output_features] "
-		"[-d feature] "
-		"[-e feature]\n"
-		"               "
-		"[-o output] "
-		"[-T mode] "
-		"[file]\n");
-	fprintf(stderr, "       lowdown "
-		"[-o output] "
-		"[output_features] "
-		"[-T mode] "
-		"[-X keyword] "
-		"[file]\n");
-	fprintf(stderr, "       lowdown-diff "
-		"[-s] "
-		"[output_features] "
-		"[-d feature] "
-		"[-e feature]\n"
-		"                    "
-		"[-o output] "
-		"[-T mode] "
-		"oldfile "
-		"[file]\n");
+	if (!diff) {
+		fprintf(stderr, 
+			"usage: lowdown [-s] [input_options] [output_options] [-M metadata]\n"
+			"               [-m metadata] [-o output] [-T mode] [-X keyword] [file]\n");
+	} else
+		fprintf(stderr, 
+			"usage: lowdown-diff [-s] [input_options] [output_options] [-M metadata]\n"
+			"                    [-m metadata] [-o output] [-T mode] oldfile [newfile]\n");
 	return 1;
 }
