@@ -1,4 +1,4 @@
-/*	$Id: term.c,v 1.45 2021/02/03 18:07:25 kristaps Exp $ */
+/*	$Id: term.c,v 1.47 2021/02/17 10:05:48 kristaps Exp $ */
 /*
  * Copyright (c) 2020--2021 Kristaps Dzonsons <kristaps@bsd.lv>
  *
@@ -147,6 +147,7 @@ static const struct sty sty_chng_del =	{ 0, 0, 0, 0, 100,  0, 0 };
  */
 
 static const struct sty sty_ddata_pfx =	{ 0, 0, 0, 0,   0, 93, 0 };
+static const struct sty sty_fdef_pfx =	{ 0, 0, 0, 0,   0, 92, 1 };
 static const struct sty sty_bkqt_pfx =	{ 0, 0, 0, 0,   0, 37, 0 };
 static const struct sty sty_oli_pfx =	{ 0, 0, 0, 0,   0, 93, 0 };
 static const struct sty sty_uli_pfx =	{ 0, 0, 0, 0,   0, 93, 0 };
@@ -553,11 +554,12 @@ rndr_buf_startline_prefixes(struct term *term,
 		rndr_buf_advance(term, 4);
 		break;
 	case LOWDOWN_FOOTNOTE_DEF:
+		rndr_node_style_apply(&sinner, &sty_fdef_pfx);
 		if (!rndr_buf_style(out, &sinner))
 			return 0;
 		pstyle = 1;
 		if (emit == 0 && !hbuf_printf
-		    (out, "%2zu. ", n->rndr_footnote_def.num))
+		    (out, "%2zu. ", n->rndr_footnote_def.num + 1))
 			return 0;
 		else if (emit != 0 && !HBUF_PUTSL(out, "    "))
 			return 0;
@@ -1230,7 +1232,7 @@ rndr(struct lowdown_buf *ob, struct lowdown_metaq *mq,
 	case LOWDOWN_FOOTNOTE_REF:
 		hbuf_truncate(p->tmp);
 		if (!hbuf_printf(p->tmp, "[%zu]", 
-		    n->rndr_footnote_ref.num))
+		    n->rndr_footnote_ref.num + 1))
 			return 0;
 		rc = rndr_buf(p, ob, n, p->tmp, NULL);
 		break;
